import setuptools

with open("requirements.txt") as f:
    required = f.read().splitlines()

if __name__ == "__main__":
    setuptools.setup(
        package_data={"grandmaster": ["py.typed"]},
        install_requires=required,
    )



from grandmaster import get_task
from grandmaster.helper import (
    get_audio,
    load_image_data_from_url,
    load_image_from_data,
    load_image_from_url,
)
from todo.mytypes import ImageQueryTypedDict, ImageZeroShotQueryTypedDict

if 1:
    # image classification
    task = get_task(
        inputs={"dataType": "IMAGE"},
        outputs={"dataType": "LABEL"},
    )
    puppy: bytes = load_image_data_from_url(
        "https://hips.hearstapps.com/hmg-prod.s3.amazonaws.com/images/dog-puppy-on-garden-royalty-free-image-1586966191.jpg"
    )
    if task is not None:
        query: ImageZeroShotQueryTypedDict = {
            "image": puppy,
            "candidate_labels": ["dog", "cat"],
        }
        out = task.inference(query)
        print(out)
    else:
        print("task not found")

if 0:
    # text classification
    task = get_task(
        inputs={"dataType": "TEXT"},
        outputs={"dataType": "LABEL", "values": ["medical", "sport"]},
    )
    out = task.inference(
        "A malignant tumor in the brain is a life-threatening condition."
    )
    print(out)

if 0:
    task = get_task(
        inputs={"dataType": "image", "representation": "car"},
        outputs={"dataType": "boundingbox", "representation": "plate"},
    )
    car = load_image_from_url(
        "https://drive.google.com/uc?id=1p9wJIqRz3W50e2f_A0D8ftla8hoXz4T5"
    )
    out = task.inference(car)
    print(out)

if 0:
    task = get_task(
        inputs={
            "dataType": "image",
        },
        outputs={"dataType": "boundingbox", "representation": "face"},
    )

    obama = load_image_from_url(
        "https://raw.githubusercontent.com/ageitgey/face_recognition/master/examples/obama.jpg"
    )

    out = task.inference(obama)
    print(out)

if 0:
    # audio classification
    # No model exists yet for zero-shot audio classification.
    pass


if 0:
    # Text generation for scientific paper
    task = get_task(
        inputs={"dataType": "prompt", "domain": "scientific"},
        outputs={"dataType": "text"},
    )
    # get_task(Input.Prompt(domain="scientific"), Output.Generated())

    out = task.inference("The Transformer architecture [START_REF]")
    print(out)

if 0:
    # Speech to text
    task = get_task(
        inputs={"dataType": "audio"},
        outputs={"dataType": "text"},
    )
    # get_task(Input.Audio(), Output.Text())

    audio = get_audio()
    print(audio.shape)

    out = task.inference(audio)
    print(out)



import re
import jinja2

from typing import Any, Dict, List, Literal, Union
from pathlib import Path


from black import format_str, FileMode

"""
def get_template(template: Literal["python", "curl"]):
    file = (
        Path(__file__).resolve().parent / "templates" / f"{template}.py.jinja2"
    ).open()
    return jinja2.Template(file.read())

class Code:
    def __init__(
        self,
        inputs: List[InputsTypedDict],
        outputs: List[OutputsTypedDict],
        isCloud: bool,
    ):

        self.inputs = [self.clean_param(i) for i in inputs]
        self.outputs = [self.clean_param(o) for o in outputs]
        self.isCloud = isCloud

    @staticmethod
    def clean_param(param: Any):
        out = {}
        if param.get("dataType") is not None:
            out["dataType"] = param["dataType"]
        if param.get("representation") is not None:
            out["representation"] = param["representation"]
        if param.get("of") is not None:
            out["representation"] = param["representation"]

        return out

    def get_python(self):
        template = get_template("python")

        x = template.render(
            inputs=self.inputs,
            outputs=self.outputs,
            examples=get_examples_url(self.inputs, self.outputs),
            isCloud=self.isCloud,
        )
        x = re.sub(r"\n\s+", "\n\n", x)
        return format_str(x, mode=FileMode(line_length=55))  # 88 default

    def get_curl(self):
        return "curl code"

    def to_json(self):
        return {"python": self.get_python(), "curl": self.get_curl()}


def get_code(
    inputs: List[InputsTypedDict], outputs: List[OutputsTypedDict], isCloud: bool
):
    return Code(inputs, outputs, isCloud).to_json()
"""


def get_samples(task):
    return [
        {
            "lang": "Python",
            "source": " mustache ... {}  ",
        },
        {
            "lang": "CURL",
            "source": "...",
        },
    ]



"""
from pathlib import Path
from typing import List, Optional
from todo.mytypes import InputsTypedDict, OutputsTypedDict

from glob import glob

from grandmaster.config import EXAMPLES_DIR


class Example:
    pass


def get_examples(
    inputs: List[InputsTypedDict], outputs: Optional[List[OutputsTypedDict]] = None
):
    examples = list((EXAMPLES_DIR / inputs[0]["dataType"].lower()).glob("*"))
    examples = [e for e in examples if e.name != ".DS_Store"]
    return examples


def get_examples_url(
    inputs: List[InputsTypedDict], outputs: Optional[List[OutputsTypedDict]] = None
):

    return [example_path_to_url(e) for e in get_examples(inputs, outputs)]


def example_path_to_url(path: Path):
    BASE = "http://localhost:3000"
    parts = path.parts
    return f"{BASE}/ex/{parts[-1]}"
"""



import io
from typing import List, Literal, get_args, Dict, Any
from PIL import Image as PILImage
from pydantic import BaseModel


class Image:
    data: bytes

    def to_pil(self) -> PILImage.Image:
        return PILImage.open(io.BytesIO(self.data))


class CandidateLabel:
    label: str


class InputImageZeroShot:
    image: Image
    candidate_labels: List[CandidateLabel]


class OutputLabel(BaseModel):
    label: str
    score: float


class TASK:
    ZERO_SHOT_IMAGE_CLASSIFICATION = "ZERO_SHOT_IMAGE_CLASSIFICATION"


OutputImageZeroShot = List[OutputLabel]


MODEL_NAME_TYPE = Literal[
    "openai/clip-vit-base-patch16",
    # "openai/clip-vit-base-patch32",
    # "openai/clip-vit-large-patch14",
    # "openai/clip-vit-large-patch14-336",
]

MODEL_NAMES: List[MODEL_NAME_TYPE] = list(get_args(MODEL_NAME_TYPE))


class TaskZeroShot(Task):
    name = TASK.ZERO_SHOT_IMAGE_CLASSIFICATION

    def __init__(self):
        self.engines: Dict[
            MODEL_NAME_TYPE, Dict[Literal["model", "processor"], Any]
        ] = {}

    def setup(self):
        from transformers import CLIPProcessor, CLIPModel

        for n in MODEL_NAMES:
            self.engines[n] = {
                "model": CLIPModel.from_pretrained(n),
                "processor": CLIPProcessor.from_pretrained(n),
            }

    def predict(
        self,
        query: InputImageZeroShot,
        model_name: MODEL_NAME_TYPE = "openai/clip-vit-base-patch16",
        model_settings={},
    ) -> OutputImageZeroShot:

        image: PILImage.Image = query.image.to_pil()

        model = self.engines[model_name]["model"]
        processor = self.engines[model_name]["processor"]

        processed = processor(
            text=query.candidate_labels,
            images=image,
            return_tensors="pt",
            padding=True,
        )

        x = model(**processed)
        results = list(x.logits_per_image.softmax(dim=1).detach().numpy()[0])

        out: List[OutputLabel] = []
        for i, r in zip(query.candidate_labels, results):
            out.append(OutputLabel(label=i, score=r.item()))
        return out



# TODO delete this file

from enum import Enum
from typing import Optional

"""
tasks = Enum(
    "text-classification",
    "image-classification",
    "face-identification",
    "face-features-identification",
    "text-to-image",
    "object-detection",
)
"""


class Grandmaster:
    def __init__(self):
        self.task = None

    def input(self, x):
        pass

    def output(self, x):
        pass

    def get_task(self, input, output, domain: Optional[any] = None):
        if (input.name == "image") and (output.name == "labels"):
            self.task = "image-classification"
            self.input = input
            self.output = output
        elif (input.name == "text") and (output.name == "labels"):
            self.task = "text-classification"
            self.input = input
            self.output = output
        elif (
            input.name == "image"
            and input.representation == "car"
            and output.name == "boundingbox"
            and output.representation == "plate"
        ):
            self.task = "plate-detection"
            self.input = input
            self.output = output
        elif (
            input.name == "image"
            and output.name == "boundingbox"
            and output.representation == "face"
        ):
            self.task = "face-recognition"
            self.input = input
            self.output = output
        elif (
            input.name == "prompt"
            and domain.name == "scientific"
            and output.name == "generated"
        ):
            self.task = "text-generation-scientific"
            self.input = input
            self.output = output
        elif input.name == "audio" and output.name == "text":
            self.task = "speech-recognition"
            self.input = input
            self.output = output

    def inference(self, x, model: Optional[str] = None):
        if self.task == "image-classification":
            from transformers import CLIPProcessor, CLIPModel

            model = CLIPModel.from_pretrained("openai/clip-vit-base-patch16")
            processor = CLIPProcessor.from_pretrained("openai/clip-vit-base-patch16")

            inputs = processor(
                text=self.output.values,
                images=x,
                return_tensors="pt",
                padding=True,
            )

            outputs = model(**inputs)
            logits_per_image = (
                outputs.logits_per_image
            )  # this is the image-text similarity score
            probs = logits_per_image.softmax(
                dim=1
            )  # we can take the softmax to get the label probabilities

            print(probs)
        if self.task == "text-classification":
            from transformers import pipeline

            classifier = pipeline(
                "zero-shot-classification", model="facebook/bart-large-mnli"
            )
            res = classifier(x, self.output.values)
            print(res)
        if self.task == "plate-detection":
            from transformers import YolosFeatureExtractor, YolosForObjectDetection

            feature_extractor = YolosFeatureExtractor.from_pretrained(
                "nickmuchi/yolos-small-rego-plates-detection"
            )
            model = YolosForObjectDetection.from_pretrained(
                "nickmuchi/yolos-small-rego-plates-detection"
            )
            inputs = feature_extractor(images=x, return_tensors="pt")
            outputs = model(**inputs)

            logits = outputs.logits
            bboxes = outputs.pred_boxes

            print(bboxes)
        if self.task == "face-recognition":
            import face_recognition
            import numpy as np

            face_locations = face_recognition.face_locations(np.array(x))
            print(face_locations)

        if self.task == "text-generation-scientific":
            from transformers import pipeline

            model = pipeline("text-generation", model="facebook/galactica-125m")
            print(model(x))

        if self.task == "speech-recognition":
            import whisper

            model = whisper.load_model("base")
            result = model.transcribe("audio.mp3")
            print(result["text"])



class YOLOPlate(Model):
    def __init__(self):
        from transformers import YolosFeatureExtractor, YolosForObjectDetection

        self.feature_extractor = YolosFeatureExtractor.from_pretrained(
            "nickmuchi/yolos-small-rego-plates-detection"
        )
        self.modelx = YolosForObjectDetection.from_pretrained(
            "nickmuchi/yolos-small-rego-plates-detection"
        )

        self.inputs: InputsTypedDict = {"dataType": "IMAGE"}
        self.outputs: OutputsTypedDict = {
            "dataType": "BOUNDINGBOX",
            "representation": "PLATE",
        }

        self.model_name = "nickmuchi/yolos-small-rego-plates-detection"

    def tasks(self):
        return [
            create_task(
                self.model_name,
                self.inputs,
                self.outputs,
                self.apply,
                self.preprocess,
                self.postprocess,
            )
        ]

    def apply(self, query: ImageQueryTypedDict) -> List[ResultBoundingBoxBoxDC]:  # TODO
        image = load_image_from_data(query["image"])
        self.feature_extractor(images=image, return_tensors="pt")
        out = self.modelx(**x)  # type: ignore
        # outputs.pred_boxes
        return []

    def postprocess(self, x: Any) -> List[ResultBoundingBoxBoxDC]:
        return x

    def preprocess(self, x):
        return x



from typing import List, Generic, TypeVar, Dict, Union, Type, TypedDict, Literal

from grandmaster.inputs import File, Image, Label, UploadFile
from grandmaster.outputs import LabelWithScore, BoundingBox
from grandmaster.examples import Puppy
from grandmaster.annotations import BaseModel
from grandmaster.Model import Model


from pydantic import Field, create_model
from pydantic.main import ModelMetaclass

from abc import ABC


class TaskTwo(ABC):
    @classmethod
    def get_models(cls):
        return cls.__subclasses__()

    name: str
    default_model_name: str

    class Inputs:
        pass

    class Outputs:
        pass

    def predict(
        self,
        input,
    ) -> Outputs:
        ...

    @classmethod
    def create_inputs_with_models(cls):

        models = cls.get_models()
        model_names_str: List[str] = [m.model_name for m in models] + ["x"]
        model_names = Literal[tuple(model_names_str)]

        return create_model(
            "InputsWithModels",
            __base__=cls.Inputs,
            model_name=(model_names, model_names_str[0]),
            task_name=(Literal[cls.name], cls.name),
        )


class ImageClassification(TaskTwo):
    name = "image-classification"

    class Inputs(BaseModel):
        image: File = Field()
        candidate_labels: List[str] = Field()

    class Outputs(BaseModel):
        labels: List[LabelWithScore]

    examples = [
        {"image": "http://localhost:3000/ex/puppy.jpeg", "labels": ["dog", "cat"]},
    ]

    default_model_name = "openai/clip-vit-large-patch14-336"


class TextClassification(TaskTwo):
    name = "text-classification"

    class Inputs(BaseModel):
        text: str = Field()

    class Outputs(BaseModel):
        labels: List[LabelWithScore]

    examples = []

    default_model_name = "X"


class FaceRecognition(TaskTwo):
    name = "face-recognition"

    class Inputs(BaseModel):
        image: File = Field(example="http://localhost:3000/ex/obama.jpeg")

    class Outputs(BaseModel):
        results: List[BoundingBox]

    default_model_name = "A"


class ObjectDetection(TaskTwo):
    name = "object-detection"

    class Inputs(BaseModel):
        image: File = Field(example="http://localhost:3000/ex/obama.jpeg")

    class Outputs(BaseModel):
        results: List[BoundingBox]

    default_model_name = "A"


all_tasks: List[Type[TaskTwo]] = [
    ImageClassification,
    # TextClassification,
    # FaceRecognition,
    # ObjectDetection,
]



import torch
from typing import Any, List
from grandmaster.Model import Model
from grandmaster.Task import Task, create_task
from grandmaster.helper import load_image_from_data

from todo.mytypes import (
    InputsTypedDict,
    OutputsTypedDict,
    PromptQueryTypedDict,
    ImageQueryTypedDict,
    InputsType,
    ResultBoundingBoxBoxDC,
    ResultBoundingBoxDC,
    ResultGeneratedTextDC,
    ResultLabelDC,
    TextZeroShotQueryTypedDict,
)


class BART(Model):
    # ZeroShotTextClassification
    def __init__(self):
        self.model_name = "facebook/bart-large-mnli"
        self.inputs: InputsTypedDict = {"dataType": "TEXT"}
        self.outputs: OutputsTypedDict = {"dataType": "LABEL"}

    def tasks(self):
        return [create_task(self.inputs, self.outputs, self.model, self.postprocess)]

    def model(self, query: TextZeroShotQueryTypedDict) -> Any:
        from transformers import pipeline

        classifier = pipeline(
            task="zero-shot-classification", model="facebook/bart-large-mnli"
        )

        results = classifier(query["text"], query["candidate_labels"].split(","))
        labels: List[str] = results["labels"]  # type: ignore
        scores: List[float] = results["scores"]  # type: ignore

        out: List[ResultLabelDC] = []
        for (l, s) in zip(labels, scores):
            out.append(ResultLabelDC(label=l, score=s))
        return out

    def postprocess(self, x: Any) -> List[ResultLabelDC]:
        return x


class Galactica(Model):
    # LM Scientific
    # TODO add 'SCIENTIFIC' domain information
    def __init__(self):
        self.inputs: InputsTypedDict = {"dataType": "PROMPT"}
        self.outputs: OutputsTypedDict = {"dataType": "COMPLETION"}
        self.model_name = "facebook/galactica-125m"

    def tasks(self):
        return [create_task(self.inputs, self.outputs, self.model, self.postprocess)]

    def model(self, query: PromptQueryTypedDict) -> ResultGeneratedTextDC:
        from transformers import pipeline

        model = pipeline(task="text-generation", model="facebook/galactica-125m")
        text = query["text"]
        generated: str = model(text)  # type: ignore
        return ResultGeneratedTextDC(text=generated)

    def postprocess(self, x: Any) -> ResultGeneratedTextDC:
        return x


class Detr(Model):
    # ZeroShotObject Detection
    def __init__(self):
        self.model_name = "facebook/detr-resnet-50"

        from transformers import DetrFeatureExtractor, DetrForObjectDetection

        self.feature_extractor = lambda: DetrFeatureExtractor.from_pretrained(
            "facebook/detr-resnet-50"
        )
        self.model = lambda: DetrForObjectDetection.from_pretrained(
            "facebook/detr-resnet-50"
        )

        self.loaded_feature_extractor = None
        self.loaded_model = None

        self.inputs: InputsTypedDict = {"dataType": "IMAGE"}
        self.outputs: OutputsTypedDict = {"dataType": "BOUNDINGBOX"}

    def tasks(self):
        return [create_task(self.inputs, self.outputs, self.apply, self.postprocess)]

    def apply(self, query: ImageQueryTypedDict) -> List[ResultBoundingBoxDC]:

        model = self.get_model()
        feature_extractor = self.get_feature_extractor()

        image = load_image_from_data(query["image"])

        inputs = feature_extractor(images=image, return_tensors="pt")
        outputs = model(**inputs)  # type: ignore
        target_sizes = torch.tensor([image.size[::-1]])
        results = feature_extractor.post_process_object_detection(
            outputs, target_sizes=target_sizes
        )[0]

        out: List[ResultBoundingBoxDC] = []

        for score, label, box in zip(
            results["scores"], results["labels"], results["boxes"]
        ):
            if score > 0.9:  # TODO
                label_name = model.config.id2label[label.item()]  # type: ignore
                box = [round(i, 2) for i in box.tolist()]
                box = ResultBoundingBoxBoxDC(
                    xmin=box[0], ymin=box[1], xmax=box[2], ymax=box[3]
                )
                score = round(score.item(), 4)
                out.append(ResultBoundingBoxDC(score=score, label=label_name, box=box))

        return out

    def get_model(self):
        if self.loaded_model:
            return self.loaded_model
        else:
            self.loaded_model = self.model()
            return self.loaded_model

    def get_feature_extractor(self):
        if self.loaded_feature_extractor:
            return self.loaded_feature_extractor
        else:
            self.loaded_feature_extractor = self.feature_extractor()
            return self.loaded_feature_extractor

    def postprocess(self, x: Any) -> List[ResultBoundingBoxDC]:
        return x



from pydantic import BaseModel
from fastapi import UploadFile

File = UploadFile


class Image(BaseModel):
    image: bytes


class Audio(BaseModel):
    audio: bytes


class Label(BaseModel):
    label: str



from typing import List
from grandmaster.inputs import Image, Label
from grandmaster.outputs import LabelWithScore
from grandmaster.tasks import ImageClassification
from grandmaster.Model import Model

from transformers import pipeline


"""
class CLIP(Model):
    # image classification
    def __init__(self):
        from transformers import CLIPProcessor, CLIPModel

        self.xmodel = lambda: CLIPModel.from_pretrained("openai/clip-vit-base-patch16")
        self.xprocessor = lambda: CLIPProcessor.from_pretrained(
            "openai/clip-vit-base-patch16"
        )

        self.model_name = "openai/clip-vit-base-patch16"

        self.inputs: List[InputsTypedDict] = [
            {"dataType": "IMAGE"},
            {"dataType": "TEXT"},
        ]
        self.outputs: List[OutputsTypedDict] = [{"dataType": "LABEL"}]

    def load(self):
        self.xmodel()
        self.xprocessor()

    def tasks(self):
        return [
            create_task(
                self.model_name,
                self.inputs,
                self.outputs,
                self.apply,
                self.preprocess,
                self.postprocess,
            )
        ]

    def apply(self, query: ImageZeroShotQueryTypedDict) -> Any:
        image = load_image_from_data(query["image"])
        processed = self.xprocessor()(
            text=query["candidate_labels"].split(","),
            images=image,
            return_tensors="pt",
            padding=True,
        )
        x = self.xmodel()(**processed)  # type: ignore
        results = list(x.logits_per_image.softmax(dim=1).detach().numpy()[0])
        out: List[ResultLabelDC] = []
        for i, r in zip(query["candidate_labels"].split(","), results):
            out.append(ResultLabelDC(label=i, score=r.item()))
        return out

    def postprocess(self, x: Any) -> List[ResultLabelDC]:
        return x

    def preprocess(self, query: ImageZeroShotQueryTypedDict) -> Any:
        return query
"""

from pydantic import BaseModel

from PIL import Image as PILImage
from fastapi import UploadFile
from io import BytesIO


def uploadfile_to_pil(uploadfile: UploadFile) -> PILImage.Image:
    # Read the contents of the file
    contents = uploadfile.file.read()

    # Create a file-like object from the bytes
    file_like = BytesIO(contents)

    # Create a PIL.Image object from the file-like object
    image = PILImage.open(file_like)

    return image


class CLIP(ImageClassification):
    model_name = "openai/clip-vit-large-patch14-336"

    def load_model(self):
        model_name = "openai/clip-vit-large-patch14-336"
        self.classifier = pipeline("zero-shot-image-classification", model=model_name)

    def predict(
        self,
        input: ImageClassification.Inputs,
    ) -> ImageClassification.Outputs:
        image_pil = uploadfile_to_pil(input.image)
        candidate_labels = (
            input.candidate_labels
        )  # [x.strip() for x in input.candidate_labels.split(",")]

        labels: List[LabelWithScore] = self.classifier(
            image_pil, candidate_labels=candidate_labels
        )

        return ImageClassification.Outputs(labels=labels)


class CLIP2(ImageClassification):
    model_name = "B"

    def load_model(self):
        model_name = "openai/clip-vit-large-patch14-336"
        self.classifier = pipeline("zero-shot-image-classification", model=model_name)

    def predict(
        self, image: Image, candidate_labels: List[Label], size: int
    ) -> List[LabelWithScore]:
        out = self.classifier(image, candidate_labels=candidate_labels)
        return []



from typing import List, Type
from grandmaster.Model import Model

# from grandmaster.models.hf.facebook.models import BART, Galactica, Detr
# from grandmaster.models.hf.openai.models import CLIP, Whisper


# from grandmaster.models.hf.facebook.models import BART
# from grandmaster.models.hf.openai.models import CLIP

# from grandmaster.models.hf.facebook.models import Detr
# from grandmaster.models.community.models import FaceRecognition, YOLOPlate

from grandmaster.models.openai.clip.predict import CLIP

all_models = [CLIP]  # , YOLOPlate, , Galactica, Whisper, ]



from typing import Any, Optional
from pydantic import BaseModel
from pathlib import Path

# from whisper.tokenizer import LANGUAGES  # TODO make our own


class LabelWithScore(BaseModel):
    label: str
    score: float


class BoundingBox(BaseModel):
    xmin: float
    xmax: float
    ymin: float
    ymax: float


class Transcription(BaseModel):
    detected_language: str
    transcription: str
    segments: Any  # TODO make it more specific
    translation: Optional[str]
    txt_file: Optional[Path]  # TODO use File
    srt_file: Optional[Path]  # TODO use one of the two.



from inspect import signature
from grandmaster.models import all_models


def validate():
    for model in all_models:
        if signature(model.predict) != signature(model.task.predict):
            print(signature(model.predict))
            print(signature(model.task.predict))
            print("\n")



"""
download the models to ./weights
wget https://openaipublic.azureedge.net/main/whisper/models/65147644a518d12f04e32d6f3b26facc3f8dd46e5390956a9424a650c0ce22b9/tiny.pt -P ./weights
wget https://openaipublic.azureedge.net/main/whisper/models/ed3a0b6b1c0edf879ad9b11b1af5a0e6ab5db9205f891f668f8b0e6c6326e34e/base.pt -P ./weights
wget https://openaipublic.azureedge.net/main/whisper/models/9ecf779972d90ba49c06d968637d720dd632c55bbf19d441fb42bf17a411e794/small.pt -P ./weights
wget https://openaipublic.azureedge.net/main/whisper/models/345ae4da62f9b3d59415adc60127b97c714f32e89e936602e85993674d08dcb1/medium.pt -P ./weights
wget https://openaipublic.azureedge.net/main/whisper/models/e4b87e7e0bf463eb8e6956e646f1e277e901512310def2c24bf0e11bd3c28e9a/large-v1.pt  -P ./weights
wget https://openaipublic.azureedge.net/main/whisper/models/81f7c96c852ee8fc832187b0132e569d6c3065a3252ed18e56effd0b6a73e524/large-v2.pt  -P ./weights
"""

from typing import Optional, Any
import torch
import numpy as np
from cog import BasePredictor, Input, Path, BaseModel

from whisper.model import Whisper, ModelDimensions
from whisper.tokenizer import LANGUAGES, TO_LANGUAGE_CODE
from whisper.utils import format_timestamp

from grandmaster.Output import Transcription


class Predictor(BasePredictor):
    def setup(self):
        self.models = {}
        for model in ["tiny", "base", "small", "medium", "large-v1", "large-v2"]:
            with open(f"weights/{model}.pt", "rb") as fp:
                checkpoint = torch.load(fp, map_location="cpu")
                dims = ModelDimensions(**checkpoint["dims"])
                self.models[model] = Whisper(dims)
                self.models[model].load_state_dict(checkpoint["model_state_dict"])

    def predict(
        self,
        tasks: {
            "transcribe": {
                "inputs": [],
                "outputs": [],
            }
        },
        audio: Path = Input(description="Audio file"),
        model: str = Input(
            default="base",
            choices=["tiny", "base", "small", "medium", "large-v1", "large-v2"],
            description="Choose a Whisper model.",
        ),
        from_language: str = Input(
            choices=list(
                sorted(LANGUAGES.keys())
                + sorted([k.title() for k in TO_LANGUAGE_CODE.keys()])
            ),
            default=None,
            description="language spoken in the audio, specify None to perform language detection",
        ),
    ) -> Transcription:
        print(f"Transcribe with {model} model")
        m = self.models[model].to("cuda")

        temperature = 0
        patience = None
        suppress_tokens = "-1"
        initial_prompt = None
        condition_on_previous_text = True
        temperature_increment_on_fallback = 0.2
        compression_ratio_threshold = 2.4
        logprob_threshold = -1.0
        no_speech_threshold = 0.6

        if temperature_increment_on_fallback is not None:
            temperature = tuple(
                np.arange(temperature, 1.0 + 1e-6, temperature_increment_on_fallback)
            )
        else:
            temperature = [temperature]

        args = {
            "language": from_language,
            "patience": patience,
            "suppress_tokens": suppress_tokens,
            "initial_prompt": initial_prompt,
            "condition_on_previous_text": condition_on_previous_text,
            "compression_ratio_threshold": compression_ratio_threshold,
            "logprob_threshold": logprob_threshold,
            "no_speech_threshold": no_speech_threshold,
        }

        result = m.transcribe(str(audio), temperature=temperature, **args)

        transcription = write_vtt(result["segments"])

        # TODO
        translate = False
        if translate:
            translation = m.transcribe(
                str(audio), task="translate", temperature=temperature, **args
            )["text"]

        return Transcription(
            segments=result["segments"],
            detected_language=LANGUAGES[result["language"]],
            transcription=transcription,
            translation=None,
        )


def write_vtt(transcript):
    result = ""
    for segment in transcript:
        result += f"{format_timestamp(segment['start'])} --> {format_timestamp(segment['end'])}\n"
        result += f"{segment['text'].strip().replace('-->', '->')}\n"
        result += "\n"
    return result


def write_srt(transcript):
    result = ""
    for i, segment in enumerate(transcript, start=1):
        result += f"{i}\n"
        result += f"{format_timestamp(segment['start'], always_include_hours=True, decimal_marker=',')} --> "
        result += f"{format_timestamp(segment['end'], always_include_hours=True, decimal_marker=',')}\n"
        result += f"{segment['text'].strip().replace('-->', '->')}\n"
        result += "\n"
    return result


"""
class Whisper(Model):
    def __init__(self):
        from whisper import load_model
        self.model_name = "openai/whisper-base"
        self.modelx = load_model("base")
        
    def model(self, query : AudioQueryTypedDict) -> :
        # TODO
        # query["audio"]
        result = self.modelx.transcribe() # type: ignore
        text : str = result["text"] # type: ignore
        return ResultsDataClass(results=ResultText(text=text))
"""



from dataclasses import dataclass

from typing import List, Optional, Type, Union

from typing import TypedDict, Literal

from typing_extensions import NotRequired

# inputs
class InputsTypedDict(TypedDict):
    dataType: Literal["TEXT", "IMAGE", "VIDEO", "AUDIO", "PROMPT"]


# outputs
class OutputsTypedDict(TypedDict):
    dataType: Literal["BOUNDINGBOX", "LABEL", "COMPLETION"]
    representation: NotRequired[Literal["FACE", "PLATE"]]


# model
class ModelTypedDict(TypedDict):
    name: str  # e.g. "bert-base-uncased"
    domain: NotRequired[Literal["SCIENTIFIC"]]


# query
class ImageQueryTypedDict(TypedDict):
    image: bytes


class ImageZeroShotQueryTypedDict(TypedDict):
    image: bytes
    candidate_labels: str  # comma separated


class TextZeroShotQueryTypedDict(TypedDict):
    text: str
    candidate_labels: str  # comma separated


class AudioQueryTypedDict(TypedDict):
    audio: bytes


class PromptQueryTypedDict(TypedDict):
    text: str


# results
# pascal_voc style


@dataclass
class ResultBoundingBoxBoxDC:
    xmin: float
    ymin: float
    xmax: float
    ymax: float


@dataclass
class ResultBoundingBoxDC:
    box: ResultBoundingBoxBoxDC
    score: Optional[float]
    label: str


@dataclass
class ResultBoundingBoxOfACarDC:
    box: ResultBoundingBoxBoxDC
    score: float
    label: str


@dataclass
class ResultLabelDC:
    label: str
    score: float


@dataclass
class ResultTextDC:
    text: str


@dataclass
class ResultGeneratedTextDC:
    text: str


ResultDC = Union[List[ResultLabelDC], List[ResultBoundingBoxDC]]
ResultType = Union[Type[List[ResultLabelDC]], Type[List[ResultBoundingBoxDC]]]


@dataclass
class ResultsDC:
    results: List[ResultDC]


# representation output


@dataclass
class RepresentationDC:
    output: Literal["PLATE"]


# visibles

InputsType = Type[InputsTypedDict]
OutputsType = Type[OutputsTypedDict]

QueryTypedDict = Union[
    ImageQueryTypedDict,
    ImageZeroShotQueryTypedDict,
    TextZeroShotQueryTypedDict,
    AudioQueryTypedDict,
]

QueryType = Union[
    Type[ImageQueryTypedDict],
    Type[ImageZeroShotQueryTypedDict],
    Type[TextZeroShotQueryTypedDict],
    Type[AudioQueryTypedDict],
]


# Task


class TaskTypedDict(TypedDict):
    inputs: InputsTypedDict
    outputs: OutputsTypedDict



from typing import Any, List
from grandmaster.Model import Model
from grandmaster.Task import create_task
from grandmaster.helper import load_image_from_data
from todo.mytypes import (
    ImageQueryTypedDict,
    InputsTypedDict,
    OutputsTypedDict,
    ResultBoundingBoxBoxDC,
    ResultBoundingBoxDC,
)


class FaceRecognition(Model):
    def __init__(self):

        self.inputs: InputsTypedDict = {"dataType": "IMAGE"}
        self.outputs: OutputsTypedDict = {
            "dataType": "BOUNDINGBOX",
            "representation": "FACE",
        }
        self.model_name = "community/face_recognition"

    def tasks(self):
        return [
            create_task(
                self.model_name,
                self.inputs,
                self.outputs,
                self.apply,
                self.preprocess,
                self.postprocess,
            )
        ]

    def apply(self, query: ImageQueryTypedDict) -> List[ResultBoundingBoxDC]:
        import face_recognition
        import numpy as np

        image = load_image_from_data(query["image"])
        boxes = face_recognition.face_locations(np.array(image))

        out: List[ResultBoundingBoxDC] = []
        for (t, r, b, l) in boxes:
            box = ResultBoundingBoxBoxDC(xmin=l, ymin=t, xmax=r, ymax=b)
            out.append(ResultBoundingBoxDC(score=None, label="face", box=box))
        return out

    def postprocess(self, x: Any) -> List[ResultBoundingBoxBoxDC]:
        return x

    def preprocess(self, x):
        return x



import inspect
from grandmaster import models
from typing import Optional, List, Literal

from pydantic import BaseModel, create_model
from grandmaster.helper import read_image

from grandmaster.tasks import all_tasks


class Task:
    def get_model(self, input) -> models.Model:
        self.task_name = input.task_name
        self.model_name = input.model_name

        for model in models.all_models:
            if model.task.name == self.task_name:
                return model()
        raise ValueError(f"Task {self.task_name} not found")

    def __call__(self, *args, **kwargs) -> models.Model:
        return self.get_model(*args, **kwargs)


from pydantic import BaseModel


# TODO generate this dynamically
class Input(BaseModel):
    task_name: str
    model_name: str


def predict_int(input: Input):
    task = Task()
    model = task.get_model(input)
    return model.prediction(input)


def xxx():
    import requests


from typing import List
from pydantic import BaseModel
from fastapi import UploadFile
import json


class Input(BaseModel):
    image: UploadFile
    candidate_labels: List[str]


import requests


import requests
from typing import List
from pydantic import BaseModel
from fastapi import UploadFile


class Input(BaseModel):
    image: UploadFile
    candidate_labels: List[str]


def fastapi_pydantic_to_request(
    pydantic_object: BaseModel, url: str
) -> requests.Response:
    files = {}

    for field, value in pydantic_object.dict().items():
        if isinstance(value, UploadFile):
            files[field] = (value.filename, value.file)
        elif isinstance(value, List):
            files[field] = (None, str(",".join(value)))
        else:
            files[field] = (None, str(value))

    response = requests.post(url, files=files)

    return response


def predict(*args, **kwargs):
    task_name = kwargs["task_name"]

    task = None
    for t in all_tasks:
        if t.name == task_name:
            task = t
            break

    if not task:
        return

    if kwargs.get("api"):
        print("Use api")

        response = fastapi_pydantic_to_request(
            task.Inputs(**kwargs), "http://localhost:8080/image-classification"
        )

        x = response.json()
        print(x)
        return x

    # TODO
    if "image" in kwargs:
        kwargs["image"] = read_image(kwargs["image"])

    models = task.get_models()

    model = None
    for m in models:
        if m.model_name == t.default_model_name:
            model = m
            break

    mm = model()
    mm.load_model()
    x = mm.predict(task.Inputs(**kwargs))
    return x.dict()



from typing import Union
from io import BytesIO
from PIL import Image
from fastapi import UploadFile
import requests


def read_image(image: Union[str, bytes, BytesIO, Image.Image]) -> UploadFile:
    """
    Reads an image from a file path, bytes, BytesIO, PIL Image object, or URL,
    and returns a FastAPI UploadFile object.
    """
    if isinstance(image, str):
        if image.startswith("http"):
            response = requests.get(image)
            contents = response.content
        else:
            with open(image, "rb") as f:
                contents = f.read()
    elif isinstance(image, bytes):
        contents = image
    elif isinstance(image, BytesIO):
        contents = image.getvalue()
    elif isinstance(image, Image.Image):
        with BytesIO() as output:
            image.save(output, format="PNG")
            contents = output.getvalue()
    else:
        raise TypeError(
            "Invalid input image type. Must be str, bytes, BytesIO, PIL Image, or URL."
        )

    binary_io = BytesIO(contents)
    return UploadFile(filename="image.png", file=binary_io)



from typing import Any, Optional, Union
from grandmaster.helper import load_image_from_url

from todo.mytypes import InputsTypedDict, QueryTypedDict


class Input:
    def __init__(self):
        self.type = "image"


class Image:
    def __init__(self, representation: Optional[str] = None):
        self.name = "image"
        self.representation = representation


class Text:
    def __init__(self):
        self.name = "text"


class Prompt:
    def __init__(self):
        self.name = "prompt"


from pathlib import Path


def load_image(image: Union[Path, str]):
    pass


def parse_query(query: Any, inputs: InputsTypedDict) -> QueryTypedDict:
    if inputs["dataType"] == "IMAGE":
        if type(query) == str:
            if "https://" in query or "http://" in query:
                return load_image_from_url(query)
    raise ValueError("query type not supported")


from pydantic import BaseModel


class Audio(BaseModel):
    file: Any  #
    language: Optional[str] = None
